/**
 * NEWS Project
 *
 * File:         AnomalyDetector.java
 * RCS:          $Id: AnomalyDetector.java,v 1.26 2008/10/25 19:19:41 drc915 Exp $
 * Description:  AnomalyDetector class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 7, 2008 at 9:47:57 AM
 * Language:     Java
 * Package:      edu.northwestern.news.anomaly
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2008, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.news.anomaly;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Random;
import java.util.Map.Entry;

import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.logging.LoggerChannel;
import org.gudy.azureus2.plugins.peers.Peer;
import org.gudy.azureus2.plugins.utils.LocaleUtilities;

import com.azureus.plugins.aznetmon.main.RSTPacketStats;

import edu.northwestern.news.MainGeneric;
import edu.northwestern.news.NEWSConfiguration;
import edu.northwestern.news.anomaly.AnomalyProperties.MovingAverageConfig;
import edu.northwestern.news.asn.NetworkAdminASNImpl;
import edu.northwestern.news.asn.NetworkAdminASNLookupImpl;
import edu.northwestern.news.dht.IDDBReadAction;
import edu.northwestern.news.dht.IDDBWriteAction;
import edu.northwestern.news.dht.IDHTOption;
import edu.northwestern.news.dht.IDistributedDatabase;
import edu.northwestern.news.dht.IDistributedDatabaseEvent;
import edu.northwestern.news.dht.azureus.IDHTWriteAction;
import edu.northwestern.news.stats.Statistics;
import edu.northwestern.news.timer.ITimer;
import edu.northwestern.news.timer.ITimerEvent;
import edu.northwestern.news.timer.ITimerEventPerformer;
import edu.northwestern.news.util.PingManager;
import edu.northwestern.news.util.PluginInterface;
import edu.northwestern.news.util.TraceRouteRunner;
import edu.northwestern.news.util.Util;
import edu.northwestern.news.Main;

/**
 * @author David Choffnes &lt;drchoffnes@cs.northwestern.edu&gt;
 *
 * The AnomalyDetector class uses various statistics and semantic 
 * information to determine if an anomaly has occurred locally. 
 * 
 * Major components:
 * 1) Calculate moving averages for various data sources
 * 2) If above threshold, publish to DHT
 * 3) If other anomalies in DHT, raise alarm
 */
public class AnomalyDetector implements ITimerEventPerformer {
	
	public class AnomalyWriteAction implements IDDBWriteAction{

//		@Override
		public void handleComplete(IDistributedDatabaseEvent event) {
			Main.getLogger().log("Write for "+event.getKey().getKey()+" complete!");
			
		}

//		@Override
		public void handleTimeout(IDistributedDatabaseEvent event) {
			Main.getLogger().log("Write for "+event.getKey().getKey()+" timed out!");
			
		}

//		@Override
		public void handleWrite(IDistributedDatabaseEvent event) {
//			Main.getLogger().log("Anomaly write for "+event.getKey().getKey()+" handled!");
			
		}
		
	}

	

	
	public class AnomalyReadAction implements IDDBReadAction{

//		@Override
		public void handleComplete(IDistributedDatabaseEvent event) {
//			Main.getLogger().log("Read complete!");
			
		}

//		@Override
		public void handleRead(byte[] b, IDistributedDatabaseEvent event) {
			ByteBuffer bBuffer = ByteBuffer.wrap(b);
			bBuffer.order(ByteOrder.LITTLE_ENDIAN);
			AnomalyRecord ar = new AnomalyRecord();
			ar.deserialize(bBuffer);
			
			String key = (String)event.getKey().getKey();
			
			if (!TESTING ){
				if (Util.currentGMTTime()-ar.timeGMT>bucketSize){
					Main.getLogger().log("Reading stale anomaly!");
					return;
				} 
				if (AnomalyRecord.getMyId()==ar.id){
					if (DEBUG) Main.getLogger().log("Reading my own value...");
					return;
				} else if (ar.id==-1){
					if (DEBUG) Main.getLogger().log("No id  for anomaly record, ignoring!");
					return;
				} else if (ar.stdDevThresh < lastStdDev){
					if (DEBUG) Main.getLogger().log("Read alert with lower sensitivity, ignoring!");
					return;
				} if (!ar.valid){
					return;
				}
			}
			
			if (DEBUG) Main.getLogger().log("Found anomaly from id " +ar.id + "(my id:"+AnomalyRecord.getMyId()+")");
			
			if (globalAnomalyCounts.get(key)==null) globalAnomalyCounts.put(key, new HashMap<Integer, AnomalyRecord>());
			globalAnomalyCounts.get(key).put(ar.id, ar);
			
			// if not just a prefix-level confirmation, use a multiplicative 
			// factor to reduce false alarms
			int sizeFactor = 1;
			if (key.startsWith("Country")) sizeFactor = 3; // country
			else if (key.startsWith("AS")) sizeFactor = 2;
			
			// get map of <Region id, text description> for alarms
			HashMap<String, Integer> foundAnomalyTypes = new HashMap<String, Integer>();
			HashMap<String, String> alarms = checkAlarms(foundAnomalyTypes);
			
			// at least one anomalous region
			if (alarms.size()>0){
				LocaleUtilities loc = Main.getPluginInterface().getUtilities().getLocaleUtilities();
				String logString = "";
				String intro = loc.getLocalisedMessageText("news.messages.alarm");
				logString += "["
				+ DateFormat.getDateTimeInstance().format(
						new Date()) + "] "+
						intro+"\n";
				for (Entry<String, String> ent : alarms.entrySet()) {
					String parts[] = ent.getKey().split("\\:");
					String region = "";
					if (parts.length==2) region = loc.getLocalisedMessageText("news.regions."+parts[0].toLowerCase()) + 
						": "+parts[1];
					else region = ent.getKey();
					logString+= loc.getLocalisedMessageText("news.misc.region")
					+ " [" + region+"], "+
					loc.getLocalisedMessageText("news.misc.anomalous")+ " " +
					loc.getLocalisedMessageText("news.misc.sensors")+" "+
					ent.getValue()+"\n";
				}
				Main.getLogger().logAlert(LoggerChannel.LT_INFORMATION, 
						logString);
				
				if (!INSERT_TEST_ANOMALIES)
					Statistics.getInstance().addAnomalyAlarm(globalAnomalyCounts.get(key).size(), ar, 
						key, foundAnomalyTypes.get(key));
				// notify listeners
				for (AnomalyListener al : listeners) 
					al.alarmRaised(key, ar, 
						globalAnomalyCounts.get(key).size());
				
				globalAnomalyCounts.get(key).clear();
			}
			
		}

		private HashMap<String, String> checkAlarms(HashMap<String, Integer> anomalyTypes) {
			
			// TODO the peers in the larger regions should be exclusive of those in narrower ones
			LocaleUtilities loc = Main.getPluginInterface().getUtilities().getLocaleUtilities();
			HashMap<String, String> toReturn = new HashMap<String, String>();
			String prefixAlarm="", asAlarm="", countryAlarm = "";
			for (Entry<String, HashMap<Integer, AnomalyRecord>> ent : 
					globalAnomalyCounts.entrySet()){
				int sizeFactor = 1; // prefix
				if (ent.getKey().startsWith("Country")) sizeFactor = 3; // country
				else if (ent.getKey().startsWith("AS")) sizeFactor = 2; // as
				
				HashMap<String, Integer> metricCounts = new HashMap<String, Integer>();
				for (String s : metricIds.keySet()) metricCounts.put(s, 0);
				// figure out if they overlap
				for (AnomalyRecord ar : ent.getValue().values()){
					for (Entry<String, Integer> ent2 : metricIds.entrySet()){
						if ((ar.anomalyTypes&ent2.getValue().intValue())>0){
							metricCounts.put(ent2.getKey(), metricCounts.get(ent2.getKey())+1);																			}
					}
				}
				
				ArrayList<String> alarmMetrics = new ArrayList<String>();
				for (Entry<String, Integer> ent2 : metricCounts.entrySet()){
					if (ent2.getValue()>alarmThreshold*sizeFactor){
						alarmMetrics.add(
								loc.getLocalisedMessageText("news.sensors."+ent2.getKey())+
								" ("+ent2.getValue()+" "+loc.getLocalisedMessageText("news.misc.cases")+")");
						if (anomalyTypes.get(ent.getKey())==null){
							anomalyTypes.put(ent.getKey(), 0);
						}
						anomalyTypes.put(ent.getKey(), anomalyTypes.get(ent.getKey())+metricIds.get(ent2.getKey()));

					}
				}
				// TODO test
				// need more than one anomalous signal!
				if (alarmMetrics.size()>1){
					if (sizeFactor==1) prefixAlarm=ent.getKey();
					else if (sizeFactor==2) asAlarm = ent.getKey();
					else countryAlarm = ent.getKey();
					
					toReturn.put(ent.getKey(), alarmMetrics.toString());
				}
			}
			
			if (prefixAlarm.length()!=0){
				if (prefixAlarmRaised){
					// suppress if alarm was already raised
					toReturn.remove(prefixAlarm);
					prefixAlarm = "";
					Main.getLogger().log("Suppressing duplicate alarm on prefix...");
				}
				prefixAlarmRaised = true;
			}
			// don't show alarm for AS if prefix isn't affected...
			if (asAlarm.length()>0 ) {
				if (!prefixAlarmRaised){
					toReturn.remove(asAlarm);
					asAlarm = "";
				} else{
					if (asAlarmRaised){
						// suppress if alarm was already raised
						toReturn.remove(asAlarm);
						asAlarm = "";
						Main.getLogger().log("Suppressing duplicate alarm on AS...");
					}
					asAlarmRaised = true;
				}
			}
			// don't show alarm for country if AS isn't affected
			if (countryAlarm.length()>0 ){
				if (!asAlarmRaised){
					toReturn.remove(countryAlarm);
					countryAlarm = "";
				} else{
					if (countryAlarmRaised){
						// suppress if alarm was already raised
						toReturn.remove(countryAlarm);
						countryAlarm = "";
						Main.getLogger().log("Suppressing duplicate alarm on country...");
					}
					countryAlarmRaised = true;
				}
			}
			
			return toReturn;
		}

//		@Override
		public void handleTimeout(IDistributedDatabaseEvent event) {
			Main.getLogger().log("Read timed out!");
			
		}
		
	}

	private static final boolean DEBUG = false;
	protected static final boolean TESTING = false;
	private static final boolean INSERT_TEST_ANOMALIES = false;

	/** minimum time between alerts from NEWS */
	private static final long ALERT_SQUELCH_TIME = 5*60*1000;

	/** map of names to moving average that are being watched */
	HashMap<String, MovingAverage> movingAverages;
	
	/** map of names to moving average that are being watched */
	HashMap<String, Sampler> samplers;
	
	/** set of names of metrics to use */
	HashSet<String> names;
	
	/** timer for controlling frequency */
	ITimer myTimer;

	private int anomCountThreshold = 2;

	private IDistributedDatabase dht;
	
	private AnomalyWriteAction awa;
	private AnomalyReadAction ara;
	
	/** window of time to consider the same "time" for anomalies (milliseconds)*/
	protected int bucketSize = 1000*1000;
	
	/** current time region for accumulating alerts */
	private long currentBucket = -1;
	/** counts per region (prefix, AS, country) */
	private HashMap<String, HashMap<Integer, AnomalyRecord>> globalAnomalyCounts;
	/** minimum number of alerts to see before raising alarm */
	public int alarmThreshold = 5;
    /** number of sequential anomalies detected */
	private int recentAnomalies = 0;
	/** sample frequency (seconds) */
	private int sampleFrequency;
	/** last time alert was shown */
	private long lastAlertTime = -1;

	private static AnomalyDetector self;
	
	private ArrayList<AnomalyListener> listeners;

	private HashMap<String, Integer> metricIds;
	protected double lastStdDev=-1;
	private boolean prefixAlarmRaised;
	private boolean asAlarmRaised;
	private boolean countryAlarmRaised;
	private boolean initComplete = false;
	
	public AnomalyDetector(ITimer anomalyTimer){	
		
		self = this;
		myTimer = anomalyTimer;
		sampleFrequency = NEWSConfiguration.getInstance().getSampleInterval();
		
		names = new HashSet<String>();
		names.add("totalDlRate");
		names.add("totalUlRate");
		names.add("perConDlRate");
		names.add("perConUlRate");
		names.add("availability-dl");
		names.add("availability-seed");
		names.add("numPeers-dl");
		names.add("numPeers-seed");
		names.add("numLeechers");
		names.add("numSeeds");
		names.add("rtt");
		names.add("loss");
		names.add("rst");
		
		metricIds = new HashMap<String, Integer>();
		metricIds.put("totalDlRate", 1);
		metricIds.put("totalUlRate", 2);
		metricIds.put("perConDlRate", 4);
		metricIds.put("perConUlRate", 8);
		metricIds.put("availability-dl", 16);
		metricIds.put("availability-seed", 32);
		metricIds.put("numPeers-dl", 64);
		metricIds.put("numPeers-seed", 128);
		metricIds.put("numLeechers", 256);
		metricIds.put("numSeeds", 512);
		metricIds.put("rtt", 1024);
		metricIds.put("loss", 2048);
		metricIds.put("rst", 4096);
		
		
		movingAverages = new HashMap<String, MovingAverage>();
		for (String name : names){
			MovingAverageConfig mac = AnomalyProperties.getInstance().getConfig(name);
			
			if (mac!=null) 
				movingAverages.put(name, new MovingAverage(mac));
		}
		
		samplers = new HashMap<String, Sampler>();
		setupSamping();
		
		awa = new AnomalyWriteAction();
		ara = new AnomalyReadAction();
		globalAnomalyCounts = new HashMap<String, HashMap<Integer, AnomalyRecord>>();
		updateCurrentBucket();
		alarmThreshold = NEWSConfiguration.getInstance().getAlarmThreshold();
		anomCountThreshold = NEWSConfiguration.getInstance().getAnomCountThreshold();
		listeners = new ArrayList<AnomalyListener>();
		initComplete = true;
	}


	private boolean updateCurrentBucket() {
		long localTime = Util.currentGMTTime();
		long myCurrentBucket = localTime - localTime%bucketSize;
		if (myCurrentBucket!=currentBucket) {			
			currentBucket = myCurrentBucket;
			// reset alarm status
			prefixAlarmRaised = false;
			asAlarmRaised = false;
			countryAlarmRaised = false;
			return true;
		} else return false;
	}


	private void setupSamping() {
		final PluginInterface pi = MainGeneric.getPluginInterface();
		for (String name : names){
			if (movingAverages.get(name)==null) continue;
			
			if (name.equals("totalDlRate")){
				
				samplers.put(name, new Sampler(movingAverages.get(name), name, pi){

					@Override
					public void sample() {
						boolean isDownloading = false;
						for (Download d : pi.getDownloadManager().getDownloads()){
							if (d.getState()==Download.ST_DOWNLOADING) isDownloading = true;
						}
						if (!isDownloading) myAverage.removeSample();
						else myAverage.addSample(pi.getDownloadManager().getStats().getDataReceiveRate());
					}});
			} else if (name.equals("totalUlRate")){
				samplers.put(name, new Sampler(movingAverages.get(name), name, pi){

					@Override
					public void sample() {
						myAverage.addSample(pi.getDownloadManager().getStats().getDataSendRate());
					}});
			} else if (name.equals("perConDlRate")){
				samplers.put(name, new Sampler(movingAverages.get(name), name, pi){

					@Override
					public void sample() {
						double sumRates =0;
						int count = 0;
						// TODO add condition only if downloading
						for (Download d : pi.getDownloadManager().getDownloads()){
							if (d.getState()!=Download.ST_DOWNLOADING) continue;
							if (d.getPeerManager()!=null){
								for (Peer p : d.getPeerManager().getPeers()){
//									if (p.isInteresting()){
										if (p.getStats().getDownloadAverage()>0){
											sumRates+=p.getStats().getDownloadAverage();
											count++;
										}
//									}
								}
							}
						}												
						if (count>0) myAverage.addSample(sumRates/count);
						else myAverage.removeSample();
					}});
			} else if (name.equals("perConUlRate")){
				samplers.put(name, new Sampler(movingAverages.get(name), name, pi){

					@Override
					public void sample() {
						double sumRates =0;
						int count = 0;
						for (Download d : pi.getDownloadManager().getDownloads()){
							if (d.getPeerManager()!=null){
								for (Peer p : d.getPeerManager().getPeers()){
//									if (p.isInterested()){
										if (p.getStats().getUploadAverage()>0){
											sumRates+=p.getStats().getUploadAverage();
											count++;
										}
//									}
								}
							}
						}												
						if (count>0) myAverage.addSample(sumRates/count);
						else myAverage.removeSample();
					}});
			} else if (name.equals("availability-dl")){
				samplers.put(name, new Sampler(movingAverages.get(name), name, pi){

					@Override
					public void sample() {
						double sumAvail =0;
						int count = 0;
						for (Download d : pi.getDownloadManager().getDownloads()){
							if (d.getState()==Download.ST_DOWNLOADING){
								sumAvail+=d.getStats().getAvailability();
								count++;
							}
						}												
						if (count>0) myAverage.addSample(sumAvail/count);
						else myAverage.removeSample();
					}});
			} else if (name.equals("availability-seed")){
				samplers.put(name, new Sampler(movingAverages.get(name), name, pi){

					@Override
					public void sample() {
						double sumAvail =0;
						int count = 0;
						for (Download d : pi.getDownloadManager().getDownloads()){
							if (d.getState()==Download.ST_SEEDING){
								sumAvail+=d.getStats().getAvailability();
								count++;
							}
						}												
						if (count>0) myAverage.addSample(sumAvail/count);
						else myAverage.removeSample();
					}});
			} else if (name.equals("numPeers-dl")){
				samplers.put(name, new Sampler(movingAverages.get(name), name, pi){

					@Override
					public void sample() {
						double sumPeer =0;
						int count = 0;
						for (Download d : pi.getDownloadManager().getDownloads()){
							if (d.getState()==Download.ST_DOWNLOADING){
								if (d.getPeerManager()==null) continue;
								sumPeer+=d.getPeerManager().getPeers().length;
								count++;
							}
						}												
						if (count>0) myAverage.addSample(sumPeer/count);
						else myAverage.removeSample();
					}});
			} else if (name.equals("numPeers-seed")){
				samplers.put(name, new Sampler(movingAverages.get(name), name, pi){

					@Override
					public void sample() {
						double sumPeer =0;
						int count = 0;
						for (Download d : pi.getDownloadManager().getDownloads()){
							if (d.getState()==Download.ST_SEEDING){
								if (d.getPeerManager()==null) continue;
								sumPeer+=d.getPeerManager().getPeers().length;
								count++;
							}
						}												
						if (count>0) myAverage.addSample(sumPeer/count);
						else myAverage.removeSample();
					}});
			} else if (name.equals("numLeechers")){
				samplers.put(name, new Sampler(movingAverages.get(name), name, pi){

					@Override
					public void sample() {
						double sumPeer =0;
						int count = 0;
						for (Download d : pi.getDownloadManager().getDownloads()){
							
							if (d.getPeerManager()==null) continue;
							sumPeer+=d.getPeerManager().getStats().getConnectedLeechers();
							count++;
							
						}												
						if (count>0) myAverage.addSample(sumPeer/count);
						else myAverage.removeSample();
					}});
			} else if (name.equals("numSeeds")){
				samplers.put(name, new Sampler(movingAverages.get(name), name, pi){

					@Override
					public void sample() {
						double sumPeer =0;
						int count = 0;
						for (Download d : pi.getDownloadManager().getDownloads()){

								if (d.getPeerManager()==null) continue;
								sumPeer+=d.getPeerManager().getStats().getConnectedSeeds();
								count++;
							
						}												
						if (count>0) myAverage.addSample(sumPeer/count);
						else myAverage.removeSample();
					}});
			} else if (name.equals("rtt")){
				samplers.put(name, new Sampler(movingAverages.get(name), name, pi){

					@Override
					public void sample() {
						myAverage.addSample(PingManager.getInstance().getRttSample());
					}});
			}  else if (name.equals("loss")){
				samplers.put(name, new Sampler(movingAverages.get(name), name, pi){

					@Override
					public void sample() {
						myAverage.addSample(TraceRouteRunner.getInstance().getLossSample());
					}});
			}  else if (name.equals("rst")){
				samplers.put(name, new Sampler(movingAverages.get(name), name, pi){

					@Override
					public void sample() {
						try {
						myAverage.addSample(RSTPacketStats.getInstance().getMostRecent().deltaConnReset);
//						System.out.println("RST, found: "+RSTPacketStats.getInstance().getMostRecent().deltaConnReset);
						} catch (Exception e){
							// do nothing
						}
					}});
			}
				
		}
		
	}


//	@Override
	public void perform(ITimerEvent event) {
		if (!initComplete) return;
		
		// check for updates
		for (String name: names){
			if (movingAverages.get(name)!=null && !movingAverages.get(name).configEquals(					
					AnomalyProperties.getInstance().getConfig(name))){
				System.out.println("New configuration for "+name+"!");
				
				movingAverages.put(name, new MovingAverage(
						AnomalyProperties.getInstance().getConfig(name)));
				samplers.get(name).setMovingAverage(movingAverages.get(name));
			}
			if (lastStdDev<0) {
				lastStdDev = AnomalyProperties.getInstance().getConfig(name).stdDevThresh;
			}
		}
		alarmThreshold = NEWSConfiguration.getInstance().getAlarmThreshold();
		anomCountThreshold = NEWSConfiguration.getInstance().getAnomCountThreshold();
		
		for (String name : names){
//			if (DEBUG) System.out.println("Sampling "+name);
			if (samplers.get(name)!=null) samplers.get(name).sample();
		}
		
		int anomalyCount = 0;
		int rateCount = 0;
		int window=0;
		ArrayList<String> anomalousNames = new ArrayList<String>();
		int anomalyTypes = 0;
		boolean shouldAnnounce = false;
		int emptyEntryCount = Integer.MAX_VALUE;
		LocaleUtilities loc = Main.getPluginInterface().getUtilities().getLocaleUtilities();
		for (String name : names){
			if (movingAverages.get(name).isAnomalous()){				
				if (name.contains("Rate")){
					// make sure a significant drop has occurred
					if (movingAverages.get(name).getDeviationPercent()<
							NEWSConfiguration.getInstance().getPctDiffTrigger()/100.0) continue;
					window = movingAverages.get(name).getWindowSize();
					lastStdDev = movingAverages.get(name).getStdDevThresh();
					rateCount++;
				}
				anomalyCount++;
				anomalousNames.add(loc.getLocalisedMessageText("news.sensors."+name));
				anomalyTypes |= metricIds.get(name);
			} else if (!movingAverages.get(name).isValid()){
				if (movingAverages.get(name).entryCount()!=0){
					if (emptyEntryCount>movingAverages.get(name).getEmptyEntryCount())
						emptyEntryCount = movingAverages.get(name).getEmptyEntryCount();
					shouldAnnounce=true;
				}
			} else if (movingAverages.get(name).isValid()) shouldAnnounce=false;
		}
		if (shouldAnnounce && emptyEntryCount<Integer.MAX_VALUE){
			String finalText = emptyEntryCount + " more " +
			"samples to go...";
			if (emptyEntryCount==1){
				finalText = "will be ready to go after next sample!";
			}
			Main.getLogger().log("["
					+ DateFormat.getDateTimeInstance().format(
							new Date()) + "] "+
					"NEWS is still warming up, "+ 
					finalText);
		}
		
		if (anomalyCount>=anomCountThreshold  && rateCount>0){
			String intro = loc.getLocalisedMessageText("news.messages.alert");
			if (canDisplayAlert()) Main.getLogger().logAlert(LoggerChannel.LT_INFORMATION, 
					"["
					+ DateFormat.getDateTimeInstance().format(
							new Date()) + "] "+
							intro+"\n\n"
					+ loc.getLocalisedMessageText("news.misc.anomalous") + 
					" " + loc.getLocalisedMessageText("news.misc.sensors")+ ": "+
					anomalousNames);
			else if (DEBUG) System.out.println("Suppressing log...");
			recentAnomalies++;
			if (dht!=null || MainGeneric.getDistributedDatabase()!=null){
				if (dht == null) dht = MainGeneric.getDistributedDatabase();
				NetworkAdminASNImpl asnInfo = NetworkAdminASNLookupImpl.getInstance().getASNInfo();
				String keys[] = new String[]{"Prefix:"+asnInfo.getBGPPrefix(), "AS:"+asnInfo.getAS(), "Country:"+asnInfo.getCountry()};
				
				AnomalyRecord ar = new AnomalyRecord();
				ar.windowSize = window;
				ar.stdDevThresh = lastStdDev;
				ar.numAnomalies = anomalyCount;
				ar.id = AnomalyRecord.getMyId();
				ar.anomalyTypes = anomalyTypes;
				Main.getLogger().log("Writing anomaly for id "+ar.id);
				Statistics.getInstance().addAnomalyEvent(ar);
				// notify listeners
				for (AnomalyListener al : listeners) al.alertRaised(ar);
				
				
				if (updateCurrentBucket()){
					for (String key : keys){
						if (globalAnomalyCounts.get(key)==null) 
							globalAnomalyCounts.put(key, new HashMap<Integer, AnomalyRecord>());
						globalAnomalyCounts.get(key).put(ar.getMyId(), ar);
					}
				}
				
				ArrayList<ByteArrayOutputStream> baoss = new ArrayList<ByteArrayOutputStream>();
				for (String key : keys){
					ByteArrayOutputStream baos = new ByteArrayOutputStream();
					try {
						ar.serialize(baos);					
						baoss.add(baos);
						dht.write(key, baoss, 
								NEWSConfiguration.getInstance().getSampleInterval(), awa);
					} catch (IOException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
				
				for (String key : keys){
					dht.read(key, "Checking for anomalies in region: "+key + " @"+Util.currentGMTTime(), 
							NEWSConfiguration.getInstance().getSampleInterval(), 
							ara, dht.getOptions().getExhaustiveRead());
				}
				
			}
		} else {
			recentAnomalies =0;
		}
		
		if (sampleFrequency!=NEWSConfiguration.getInstance().getSampleInterval()){			
			if (DEBUG) System.out.println("Setting sample frequency to "+
					NEWSConfiguration.getInstance().getSampleInterval() + 
					" seconds!");
			event.cancel();
			sampleFrequency = NEWSConfiguration.getInstance().getSampleInterval();
			myTimer.addPeriodicEvent(sampleFrequency*1000, this);
		}
		
		if (INSERT_TEST_ANOMALIES) doTestAnomalies();
		
	} // end perform
	
	private void doTestAnomalies() {
		int NUM_REPS = 10;
		
		for (int i = 0; i < NUM_REPS; i++){
		if (dht!=null || MainGeneric.getDistributedDatabase()!=null){
			if (dht == null) dht = MainGeneric.getDistributedDatabase();
			NetworkAdminASNImpl asnInfo = NetworkAdminASNLookupImpl.getInstance().getASNInfo();
		String keys[] = new String[]{"Prefix:"+asnInfo.getBGPPrefix(), "AS:"+asnInfo.getAS(), "Country:"+asnInfo.getCountry()};
		
		AnomalyRecord ar = new AnomalyRecord();
		ar.windowSize = 20;
		ar.stdDevThresh = lastStdDev;
		ar.numAnomalies = 3;
		ar.id = (int)(Math.random()*Integer.MAX_VALUE);
		ar.anomalyTypes = Integer.MAX_VALUE;
		Main.getLogger().log("Writing anomaly for id "+ar.id);
//		Statistics.getInstance().addAnomalyEvent(ar);
		// notify listeners
//		for (AnomalyListener al : listeners) al.alertRaised(ar);
		
		if (updateCurrentBucket()){
			for (String key : keys){
				if (globalAnomalyCounts.get(key)==null) 
					globalAnomalyCounts.put(key, new HashMap<Integer, AnomalyRecord>());
				globalAnomalyCounts.get(key).put(ar.getMyId(), ar);
			}
		}
		
		ArrayList<ByteArrayOutputStream> baoss = new ArrayList<ByteArrayOutputStream>();
		for (String key : keys){
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			try {
				ar.serialize(baos);					
				baoss.add(baos);
				dht.write(key, baoss, 
						NEWSConfiguration.getInstance().getSampleInterval(), awa);
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		for (String key : keys){
			dht.read(key, "Checking for anomalies in region: "+key + " @"+Util.currentGMTTime(), 
					NEWSConfiguration.getInstance().getSampleInterval(), 
					ara, dht.getOptions().getExhaustiveRead());
		}
		}
		} // end loop of repetitions
		
	}


	private boolean canDisplayAlert(){
		if (!NEWSConfiguration.getInstance().isSuppressWarnings() && 
				Util.currentGMTTime()-lastAlertTime > ALERT_SQUELCH_TIME){
			lastAlertTime = Util.currentGMTTime();
			return true;
		}
		return false;
		
	}
	
	public static AnomalyDetector getInstance(){
		return self;
	}
	
	public void addListener(AnomalyListener al){
		listeners.add(al);
	}
	
	public void removeListener(AnomalyListener al){
		listeners.remove(al);
	}

}
