/**
 * NEWS Project
 *
 * File:         MovingAverage.java
 * RCS:          $Id: MovingAverage.java,v 1.9 2008/08/27 12:48:17 drc915 Exp $
 * Description:  MovingAverage class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 7, 2008 at 9:52:28 AM
 * Language:     Java
 * Package:      edu.northwestern.news.anomaly
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2008, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.news.anomaly;

import java.util.ArrayList;

import edu.northwestern.news.anomaly.AnomalyProperties.MovingAverageConfig;

/**
 * @author David Choffnes &lt;drchoffnes@cs.northwestern.edu&gt;
 *
 * The MovingAverage class ...
 */
public class MovingAverage {
	
	private static final boolean DEBUG = false;
	/** size of the moving average window */
	private int windowSize;
	/** threshold for anomaly detection */
	private double stdDevThresh;
	/** whether signal is in anomalous state */
	private boolean isAnomalous;
	/** values that are sampled */
	private ArrayList<Double> samples;
	/** values that are sampled */
	private ArrayList<Double> samplesAnomalous;
	/** mean */
	private double mean;
	/** variance */
	private double variance;
	/** sign of the change in value that is anomalous, 0 for any direction */
	private int anomalyDirection;
	private double deviationPct;
	
	public MovingAverage(int windowSize, double stdDevThresh, 
			int anomalyDirection ){
		this.windowSize = windowSize;
		this.stdDevThresh = stdDevThresh;
		samples = new ArrayList<Double>(windowSize);
		samplesAnomalous = new ArrayList<Double>(windowSize);
		isAnomalous = false;
		mean = 0;
		variance = 0;
		this.anomalyDirection = anomalyDirection;
	}
	
	public MovingAverage(MovingAverageConfig config) {
		this.windowSize = config.windowSize;
		this.stdDevThresh = config.stdDevThresh;
		samples = new ArrayList<Double>(windowSize);
		samplesAnomalous = new ArrayList<Double>(windowSize);
		isAnomalous = false;
		mean = 0;
		variance = 0;
		this.anomalyDirection = config.anomalyDirection;
	}

	public void addSample(double value){
		if (DEBUG) System.out.println("Number of samples: "+samples.size());
		if (samples.size()==windowSize){
			if (DEBUG) System.out.println("Value: "+value+
					" Mean: "+mean + " StdDev:"+Math.sqrt(variance));
			
			if (Math.abs(value-mean) > stdDevThresh*Math.sqrt(variance)){
				
				// only signal anomaly if in proper direction
				if (anomalyDirection!=0 && 
						Math.signum(value-mean)==anomalyDirection){
					// not only variation, but threshold drop in percent
					deviationPct = Math.abs(value-mean)/mean;
					isAnomalous = true;
					if (DEBUG) System.out.println("Anomalous data!");
				}
				
				samplesAnomalous.add(value);
				if (samplesAnomalous.size()>windowSize/2){
					samples.clear();
					samples.addAll(samplesAnomalous);
					samplesAnomalous.clear();
				}
				
				return;
			} else {
				isAnomalous = false;
			}
		} else isAnomalous = false;
		
		if (samples.size()==windowSize) samples.remove(0);
		samples.add(value);

		updateStats();
	}
	
	private void updateStats(){
		mean = 0;
		for (double d : samples) mean+=d;
		mean/= samples.size();
		
		variance =0;
		for (double d : samples) variance += (d-mean)*(d-mean);
		variance/=samples.size();
		
	}
	
	public boolean isAnomalous(){
		return isAnomalous;
	}

	public int getWindowSize() {
		return windowSize;
	}

	public double getStdDevThresh() {
		return stdDevThresh;
	}

	public boolean configEquals(MovingAverageConfig config) {
		return (config.anomalyDirection==anomalyDirection && 
				config.stdDevThresh == stdDevThresh && 
				config.windowSize== windowSize);
	}

	public void removeSample() {
		if (samples.size()>0) {
			samples.remove(0);		
		}
		isAnomalous = false; // because we don't want to leave it
							// in an anomalous state when no data is coming
	}

	public boolean isValid() {		
		return samples.size()>=windowSize;
	}

	public int getEmptyEntryCount() {
		return windowSize-samples.size();
	}
	
	public int entryCount(){
		return samples.size();
	}

	public double getDeviationPercent() {
		return deviationPct;
	}
}
