/**
 * News Project
 *
 * File:         NewsPeerManager.java
 * RCS:          $Id: NewsPeerManager.java,v 1.2 2008/05/14 14:24:13 drc915 Exp $
 * Description:  NewsPeerManager class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Apr 20, 2007 at 11:23:26 AM
 * Language:     Java
 * Package:      edu.northwestern.news.position
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2007, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.news.stats;

import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.peers.Peer;

import edu.northwestern.news.MainGeneric;
import edu.northwestern.news.NEWSConfiguration;
import edu.northwestern.news.timer.ITimer;
import edu.northwestern.news.timer.ITimerEvent;
import edu.northwestern.news.timer.ITimerEventPerformer;
import edu.northwestern.news.util.HashMapCache;
import edu.northwestern.news.util.Util;

/**
 * @author David Choffnes &lt;drchoffnes@cs.northwestern.edu&gt;
 *
 * The NewsPeerManager class ...
 */
public class NewsPeerManager {
	
	/** closeness states */
	public static byte CS_UNKNOWN = 0;
	public static byte CS_CLOSE = 1;
	public static byte CS_NOT_CLOSE = -1;
	
    /** the duration that an edge-server ratio is valid */
	private static final long RATIO_VALIDITY_INTERVAL = 12*60*60*1000;
	
	/** peer source */
	public static final byte SOURCE_DHT = 1;
	public static final byte SOURCE_VIVALDI = 2;
	public static final byte SOURCE_ANNOUNCE = 3;
	public static final byte SOURCE_PEER_ADDED = 4;
	
	/** transfer status */
	public static final String NOT_CONNECTED = "NC";
	public static final String CANT_COMPARE = "CC";
	public static final String NOT_UPLOADING = "NU";
	public static final String NOT_DOWNLOADING = "ND";
	public static final String NOT_APPLICABLE = "NA";
	
	public static final int MAX_PEERS_TO_DIG = 100;
	private static final boolean VERBOSE = false;
	public static final String NOT_INTERESTING = "NID";
	public static final String NOT_INTERESTED = "NIU";
	
	public static class NewsPeer implements ITimerEventPerformer {
		String ip;
		int port;
		HashMap<Object, Boolean> keys;
		boolean getRatios;
		
		
		/** whether a node is close or not */
		byte closenessState;
		
		/** what component found this peer */
		byte source; 
		boolean isInDHT = false;
		private boolean findCluster;
		public boolean dhtLookupAttempted;
		private boolean digFailed = false;
		public short protocol = 1;
		private HashMap<String, Double> cosSimMappings;

		
		public NewsPeer(String ip, int port, Object key, boolean getRatios){
			this.ip = ip;
			this.port = port;
			keys = new HashMap<Object, Boolean>();
			keys.put(key, false);
			this.getRatios = getRatios;
			closenessState = 0;
			cosSimMappings = new HashMap<String, Double>();
			
		}

		@Override
		public boolean equals(Object obj) {
			if (obj==null) return false;
			if (obj instanceof NewsPeer){
				return ((NewsPeer)obj).ip.equals(ip);
			}
			return false;
		}

		@Override
		public int hashCode() {
			return ip.hashCode();
		}

		public void addKey(Object key) {
			synchronized (keys){
				if (!keys.containsKey(key)) keys.put(key, false);
			}
		}

		public void updatePort(int port) {
			if (port>0 && this.port <=0)
				this.port = port;
			
		}

		public void updateGetRatios(boolean getRatios) {
			
			this.getRatios = getRatios;
			
		}


		public void registerPeerSource(byte source) {
			this.source = source;
			
		}

		public void digFailed() {
			digFailed  = true;
			
		}

		public boolean shouldDig() {			
			return getRatios && closenessState >= 0 && !digFailed;
		}

		public String getIp() {
			return ip;
		}

		public int getPort() {
//			if (!MainGeneric.isAzureus()){
//				return MainGeneric.getPeerPort(ip);
//			}
			
			if (port <=1){
				if (keys.size()>0){
					for (Object key : keys.keySet()){
						if (key instanceof Download){
							Download dl = (Download)key;					
							if (dl.getPeerManager()==null) continue;
							Peer peers[] = dl.getPeerManager().getPeers(ip);
							for (Peer p : peers){
								if (p.getTCPListenPort()>0){
									port = p.getTCPListenPort();
									return port;
								}
							}
						} 
					}
				}
			}
			return port;
		}

		public short getProtocol() {
			return protocol;
		}

		

		
		
		@Override
		public String toString() {
			StringBuffer sb = new StringBuffer();
			sb.append("Ip: ");
			sb.append(ip);
			sb.append("\n");
			
			sb.append("Port: ");
			sb.append(port);
			sb.append("\n");

			
			
			return sb.toString();
		}

		/**
		 * Called as a periodic timeout
		 * @param event
		 */
		public void perform(ITimerEvent event) {
			if (MainGeneric.isShuttingDown()){
				event.cancel();
			}
			else if (cosSimMappings.size()==0 || digFailed ){
				NewsPeerManager.getInstance().remove(ip);
				event.cancel();
			} 
		}


		/**
		 * Copies the argument's data into this data
		 * @param peer the data to copy
		 */
		public void copy(NewsPeer peer) {
			this.port = peer.getPort();
			this.source = peer.source;
			this.closenessState = peer.closenessState;
			this.cosSimMappings = peer.cosSimMappings;
			this.keys = peer.keys;						
		}

		public String getRelativeDownloadPerformance() {
			String toReturn = null;
			if (closenessState < CS_CLOSE){
				toReturn =  NOT_APPLICABLE;
			} else {
				if (keys.size()==0){
					toReturn =  NOT_APPLICABLE;
				}
				Download toAdd = null;
				for (Object o : keys.keySet()){
					if (o instanceof Download){
						float sum = 0f;
						int count=0;	
						if (((Download)o).getPeerManager()==null){
							toAdd = MainGeneric.getPluginInterface().getDownloadManager().
							getDownload(((Download)o).getTorrent());
							break;
						}
						for (Peer p : ((Download)o).getPeerManager().getPeers()){
							if (NewsPeerManager.getInstance().getNewsPeer(p.getIp())==null) continue;
							if (NewsPeerManager.getInstance().getNewsPeer(p.getIp()).closenessState<1 
									&& p.getStats().getDownloadAverage()>1000){
								sum+=p.getStats().getDownloadAverage();
								count++;
							}
							
						}

						
						Peer ps[] = ((Download)o).getPeerManager().getPeers(getIp());
						if (ps.length>0){
							if (ps[0].getStats().getDownloadAverage()<1000){
								if (!ps[0].isInteresting()){
									toReturn =  NOT_INTERESTING;
								}
								else toReturn =  NOT_DOWNLOADING;
								break;
							}
							if (count==0 || sum==0) continue;
							float otherAverage = sum/count;
							
							return ((100*(ps[0].getStats().getDownloadAverage()/otherAverage))+"%");
						}
						toReturn =  NOT_CONNECTED;
					}
				}			
				if (toAdd!=null){
					Boolean b = keys.remove(toAdd);
					keys.put(toAdd, b);
				}
			}

			if (toReturn ==null) toReturn = CANT_COMPARE;
			return toReturn;
		}

		public String getRelativeUploadPerformance() {
			String toReturn = null;
			if (closenessState < CS_CLOSE){
				toReturn = NOT_APPLICABLE;
			} else {
				if (keys.size()==0){
					toReturn = NOT_APPLICABLE;
				}
				Download toAdd = null;
				for (Object o : keys.keySet()){
					if (o instanceof Download){
						float sum = 0f;
						int count=0;
						if (((Download)o).getPeerManager()==null){
							toAdd = MainGeneric.getPluginInterface().getDownloadManager().
								getDownload(((Download)o).getTorrent());
							break;
						}
						for (Peer p : ((Download)o).getPeerManager().getPeers()){
							if (NewsPeerManager.getInstance().getNewsPeer(p.getIp())==null) continue;
							if (NewsPeerManager.getInstance().getNewsPeer(p.getIp()).closenessState<1 
									&& p.getStats().getUploadAverage()>1000){
								sum+=p.getStats().getUploadAverage();
								count++;
							}
							
						}
						
						Peer ps[] = ((Download)o).getPeerManager().getPeers(getIp());
						if (ps.length>0){
							if (ps[0].getStats().getUploadAverage()<1000){
								if (!ps[0].isInteresting()){
									toReturn = NOT_INTERESTED;
								}
								else toReturn = NOT_UPLOADING;
							}
							if (count==0 || sum==0) continue;
							float otherAverage = sum/count;
							if (ps[0].getStats().getUploadAverage() == 0) continue;
							return ((100*(ps[0].getStats().getUploadAverage()/otherAverage))+"%");
						}
						toReturn = NOT_CONNECTED;
					}
				}	
				if (toAdd!=null){
					Boolean b = keys.remove(toAdd);
					keys.put(toAdd, b);
				}
			}

			if (toReturn ==null) toReturn = CANT_COMPARE;
			return toReturn;
		}

		public String getToolTipText(String text) {
			if (text.equals(CANT_COMPARE)){
				return "No other peers to compare with";
			} else if (text.equals(NOT_CONNECTED)){
				return "Not currently connected to this peer";
			}  else if (text.equals(NOT_UPLOADING)){
				return "Not uploading to this peer";
			}  else if (text.equals(NOT_INTERESTING)){
				return "Peer doesn't have data that you want";
			}  else if (text.equals(NOT_INTERESTED)){
				return "You don't have data that peer wants";
			}   else if (text.equals(NOT_DOWNLOADING)){
				return "Not downloading from this peer";
			} else if (text.equals(NOT_APPLICABLE)){
				return "No download to observe";
			}
			
			return text + " better than the average for \"not close\" peers";
		}
		
		
		
	}
	
	private static NewsPeerManager self;
	
	private HashMap<String, NewsPeer> peers;
	private String myIp;
	private ITimer timeoutTimer;
	
	public NewsPeerManager(){
		peers = new HashMap<String, NewsPeer>();
		if (MainGeneric.getPublicIpAddress()!=null){
			peers.put(MainGeneric.getPublicIpAddress(), 
					new NewsPeer(MainGeneric.getPublicIpAddress(), 0, null, true));
		}
		timeoutTimer = MainGeneric.createTimer("NewsPeerManagerTimeout");		
	}
	
	public void remove(String ip) {
		synchronized (peers){
			peers.remove(ip);
		}
		
	}

	public static synchronized NewsPeerManager getInstance(){
		if (self==null){
			self = new NewsPeerManager();
		}
		return self;
	}
	
	public void addPeer(String ip, int port, Object key, boolean getRatios){
		
		synchronized (peers){
			if (peers.get(ip)!=null){
				peers.get(ip).addKey(key);
				peers.get(ip).updatePort(port);
				peers.get(ip).updateGetRatios(getRatios); 
				// TODO decide to dig based on a closeness and previous state
			} else {
				peers.put(ip, new NewsPeer(ip, port, key, getRatios));	

				timeoutTimer.addPeriodicEvent(60*60*1000, 
					peers.get(ip));
			}
		}
		
		
	}
	
	public void updatePeerCloseness(String ip, byte closenessState){
		if (peers.get(ip)!=null){
			peers.get(ip).closenessState = closenessState;
		} else {
			System.err.println("Unknown peer!");
		}
	}


	public void registerPeerSource(String peerIp, byte source) {
		if (peers.get(peerIp)!=null){
			peers.get(peerIp).registerPeerSource(source);
		} else {
			throw new RuntimeException("Need to register peer ip before " +
					"registering source!");
		}
		
	}


	public void digFailed(String peerIp) {
		if (peers.get(peerIp)!=null){
			peers.get(peerIp).digFailed();
		} else {
			if (VERBOSE) System.err.println("Need to register peer ip before notifying of dig failure!");
		}
		
	}

	public HashSet<String> getIpsToDig() {
		HashSet<String> toReturn = new HashSet<String>();
		synchronized (peers){
			for (Entry<String, NewsPeer> ent: peers.entrySet()){
				if (ent.getValue().shouldDig()){
					toReturn.add(ent.getKey());
				}
				if (toReturn.size()>= MAX_PEERS_TO_DIG) break;
			}
		}
		return toReturn;
	}

	public void foundInDht(String peerIp) {
		if (peers.get(peerIp)!=null){
			peers.get(peerIp).isInDHT = true;
		} else {
			throw new RuntimeException("Need to register peer ip before notifying of entry in DHT!");
		}
		
	}

	public NewsPeer getNewsPeer(String peerIp) {
		if (peers.get(peerIp)==null){
//			System.err.println("Warning: Accessing peer ip not in manager!");
			return null;
		}
		return peers.get(peerIp);
	}

	public void setPeerProtocol(String peerIp, short protocol) {
		if (peers.get(peerIp)!=null){
			peers.get(peerIp).protocol = protocol;
		} else {
			throw new RuntimeException("Need to register peer ip before setting protocol!");
		}
		
	}

	
	private double classCSubnetContainsKey(Map<String, Double> edgePings, String subnet) {
		HashSet<String> keys = new HashSet<String>();
		keys.addAll(edgePings.keySet());
		for (String key : keys){
			if (key.equals(subnet) || Util.getClassCSubnet(key).equals(subnet)){
				return edgePings.get(key);
			}
		}
		return -1;
	}

	public void addPeer(NewsPeer op) {
		peers.put(op.getIp(), op);
		
	}

	public void stop() {
		if (timeoutTimer!=null){
			timeoutTimer.addEvent(System.currentTimeMillis()+100, new ITimerEventPerformer(){

				public void perform(ITimerEvent event) {
					// TODO Auto-generated method stub
					
				}});
			timeoutTimer.destroy();
		}
		self = null;
		
	}
	
	
	
}
